#include "./USB/usb.h"
#include "HardwareProfile.h"
#include "./USB/usb_function_hid.h"
#include "./USB/usb_function_cdc.h"

__CONFIG(FOSC_INTOSC & WDTE_OFF & PWRTE_ON & MCLRE_ON & CP_OFF & BOREN_ON & CLKOUTEN_OFF & IESO_OFF & FCMEN_OFF);
__CONFIG(WRT_OFF & CPUDIV_NOCLKDIV & USBLSCLK_48MHz & PLLMULT_3x & PLLEN_ENABLED & STVREN_ON & BORV_LO & LPBOR_OFF & LVP_OFF);

unsigned char ReceivedDataBuffer[64] @0x2050; // RX_DATA_BUFFER_ADDRESS
unsigned char ToSendDataBuffer[64] @0x20A0; // TX_DATA_BUFFER_ADDRESS;

USB_HANDLE USBOutHandle = 0; //USB handle.  Must be initialized to 0 at startup.
USB_HANDLE USBInHandle = 0; //USB handle.  Must be initialized to 0 at startup.
BOOL blinkStatusValid = TRUE;

char USB_Out_Buffer[CDC_DATA_OUT_EP_SIZE];
char RS232_Out_Data[CDC_DATA_IN_EP_SIZE];

unsigned char NextUSBOut;
unsigned char LastRS232Out; // Number of characters in the buffer
unsigned char RS232cp; // current position within the buffer
unsigned char RS232_Out_Data_Rdy = 0;
USB_HANDLE lastTransmission;
char USB_In_Buffer[64];
char USB_Out_Buffer[64];

/** PRIVATE PROTOTYPES *********************************************/
void BlinkUSBStatus(void);
static void InitializeSystem(void);
void ProcessIO(void);
void UserInit(void);
void YourHighPriorityISRCode(void);
void YourLowPriorityISRCode(void);
WORD_VAL ReadPOT(void);
void InitializeUSART(void);
void putcUSART(char c);
unsigned char getcUSART(void);
char IdleI2C(void);
char StartI2C(void);
char StopI2C(void);
char ReStartI2C(void);
char NAckI2C(void);
char WriteI2C(unsigned char Data);
char ReadI2C(unsigned char * Data);
char I2C_EE16BitsByteWirte(unsigned char SalveAddr, unsigned int Addr, unsigned char Data);
char I2C_EE16BitsRandomRead(unsigned char SalveAddr, unsigned int Addr, unsigned char * Data);

void interrupt ISRCode()
{
    #if defined(USB_INTERRUPT)
    USBDeviceTasks();
    #endif
}

int main(void)
{
    InitializeSystem();

    #if defined(USB_INTERRUPT)
    USBDeviceAttach();
    #endif

    while (1)
    {
        #if defined(USB_POLLING)
        USBDeviceTasks();
        #endif
        ProcessIO();
    }
}

static void InitializeSystem(void)
{
    ANSELA = 0;
    ANSELC = 0;

    OSCCONbits.IRCF = 0x0f; //16MHz HFINTOSC with 3x PLL enabled (48MHz operation)
    OSCCONbits.SCS = 0x00;
    OSCCONbits.SPLLEN = 1;
    OSCCONbits.SPLLMULT = 1;
    ACTCONbits.ACTSRC = 1; //Enable active clock tuning with USB
    ACTCONbits.ACTEN = 1;

    #if defined(USE_USB_BUS_SENSE_IO)
    tris_usb_bus_sense = INPUT_PIN; // See HardwareProfile.h
    #endif

    #if defined(USE_SELF_POWER_SENSE_IO)
    tris_self_power = INPUT_PIN; // See HardwareProfile.h
    #endif

    UserInit();

    USBDeviceInit();
}

void UserInit(void)
{
    unsigned char i;

    InitializeUSART();

    for (i = 0; i<sizeof (USB_Out_Buffer); i++)
    {
        USB_Out_Buffer[i] = 0;
    }
    NextUSBOut = 0;
    LastRS232Out = 0;
    lastTransmission = 0;

    SSPCON1bits.SSPM = 0x08; // Master mode, Clock = FOSC / (4 * (SSPADD+1))
    SSPADDbits.SSPADD = 0x27;
    SSPCON1bits.SSPEN = 1;

    ANSELCbits.ANSC0 = 0;
    ANSELCbits.ANSC1 = 0;

    mInitAllLEDs();
    mInitAllSwitches();
    mInitPOT();

    blinkStatusValid = TRUE;
}//end UserInit

void ProcessIO(void)
{
    //Blink the LEDs according to the USB device status
    if (blinkStatusValid)
        BlinkUSBStatus();

    // User Application USB tasks
    if ((USBDeviceState < CONFIGURED_STATE) || (USBSuspendControl == 1)) return;

    blinkStatusValid = FALSE;

    // HID Class
    if (!HIDRxHandleBusy(USBOutHandle))
    {
        switch (ReceivedDataBuffer[0])
        {
            case 0x80:
                mLED_1_Toggle();
                break;

            case 0x81:
                mLED_2_Toggle();
                break;

            case 0x82:
            {
                WORD_VAL w;

                w = ReadPOT();
                ToSendDataBuffer[0] = 0x82; // Internal Data Package ID
                ToSendDataBuffer[1] = 0xAA;
                ToSendDataBuffer[2] = 0x00; // User Define Code1
                ToSendDataBuffer[3] = 0x00; // User Define Code2
                ToSendDataBuffer[4] = w.v[0]; //Measured analog voltage LSB
                ToSendDataBuffer[5] = w.v[1]; //Measured analog voltage MSB
                if (!HIDTxHandleBusy(USBInHandle))
                {
                    USBInHandle = HIDTxPacket(HID_EP, (BYTE*) & ToSendDataBuffer[0], 64);
                }
                break;
            }
            case 0x83:
                ToSendDataBuffer[0] = 0x83; // Internal Data Package ID
                ToSendDataBuffer[1] = 0xAA;
                ToSendDataBuffer[2] = 0x00; // User Define Code1
                ToSendDataBuffer[3] = 0x00; // User Define Code2
                ToSendDataBuffer[4] = !sw1;

                if (!HIDTxHandleBusy(USBInHandle))
                {
                    USBInHandle = HIDTxPacket(HID_EP, (BYTE*) & ToSendDataBuffer[0], 64);
                }
                break;

            case 0xA0:
                // I2C Byte Wirte
                if (I2C_EE16BitsByteWirte(ReceivedDataBuffer[4], ((long) ReceivedDataBuffer[6] << 8) + ReceivedDataBuffer[5], ReceivedDataBuffer[7]))
                    ToSendDataBuffer[1] = 0x55; // I2C Error Code
                else
                    ToSendDataBuffer[1] = 0xAA; // I2C Tx/Rx Completed

                ToSendDataBuffer[0] = ReceivedDataBuffer[0]; // Command echo
                ToSendDataBuffer[2] = ReceivedDataBuffer[2]; // User Define Code1
                ToSendDataBuffer[3] = ReceivedDataBuffer[3]; // User Define Code2
                ToSendDataBuffer[4] = 0x00; // Rerutn data

                if (!HIDTxHandleBusy(USBInHandle))
                {
                    USBInHandle = HIDTxPacket(HID_EP, (BYTE*) & ToSendDataBuffer[0], 64);
                }
                break;

            case 0xA1:
                // I2C Random Read
                if (I2C_EE16BitsRandomRead(ReceivedDataBuffer[4], ((long) ReceivedDataBuffer[6] << 8) + ReceivedDataBuffer[5], &ToSendDataBuffer[4]))
                    ToSendDataBuffer[1] = 0x55; // I2C Error Code
                else
                    ToSendDataBuffer[1] = 0xAA; // I2C Tx/Rx Completed

                ToSendDataBuffer[0] = ReceivedDataBuffer[0]; // Command echo
                ToSendDataBuffer[2] = ReceivedDataBuffer[2]; // User Define Code1
                ToSendDataBuffer[3] = ReceivedDataBuffer[3]; // User Define Code2
                //ToSendDataBuffer[4] = 0x00; // Rerutn data

                if (!HIDTxHandleBusy(USBInHandle))
                {
                    USBInHandle = HIDTxPacket(HID_EP, (BYTE*) & ToSendDataBuffer[0], 64);
                }
                break;

            default:
                break;
        }
        //Prepare the USB module to send the data packet to the host
        USBOutHandle = HIDRxPacket(HID_EP, (BYTE*) & ReceivedDataBuffer, 64);
    }

    // CDC Class
    if (RS232_Out_Data_Rdy == 0) // only check for new USB buffer if the old RS232 buffer is
    { // empty.  This will cause additional USB packets to be NAK'd
        LastRS232Out = getsUSBUSART(RS232_Out_Data, 64); //until the buffer is free.
        if (LastRS232Out > 0)
        {
            RS232_Out_Data_Rdy = 1; // signal buffer full
            RS232cp = 0; // Reset the current position
        }
    }

    //Check if one or more bytes are waiting in the physical UART transmit
    //queue.  If so, send it out the UART TX pin.
    if (RS232_Out_Data_Rdy && mTxRdyUSART())
    {
        #if defined(USB_CDC_SUPPORT_HARDWARE_FLOW_CONTROL)
        //Make sure the receiving UART device is ready to receive data before
        //actually sending it.
        if (UART_CTS == USB_CDC_CTS_ACTIVE_LEVEL)
        {
            putcUSART(RS232_Out_Data[RS232cp]);
            ++RS232cp;
            if (RS232cp == LastRS232Out)
                RS232_Out_Data_Rdy = 0;
        }
        #else
        //Hardware flow control not being used.  Just send the data.
        putcUSART(RS232_Out_Data[RS232cp]);
        ++RS232cp;
        if (RS232cp == LastRS232Out)
            RS232_Out_Data_Rdy = 0;
        #endif
    }

    //Check if we received a character over the physical UART, and we need
    //to buffer it up for eventual transmission to the USB host.
    if (mDataRdyUSART() && (NextUSBOut < (CDC_DATA_OUT_EP_SIZE - 1)))
    {
        USB_Out_Buffer[NextUSBOut] = getcUSART();
        ++NextUSBOut;
        USB_Out_Buffer[NextUSBOut] = 0;
    }

    #if defined(USB_CDC_SUPPORT_HARDWARE_FLOW_CONTROL)
    //Drive RTS pin, to let UART device attached know if it is allowed to
    //send more data or not.  If the receive buffer is almost full, we
    //deassert RTS.
    if (NextUSBOut <= (CDC_DATA_OUT_EP_SIZE - 5u))
    {
        UART_RTS = USB_CDC_RTS_ACTIVE_LEVEL;
    }
    else
    {
        UART_RTS = (USB_CDC_RTS_ACTIVE_LEVEL ^ 1);
    }
    #endif

    //Check if any bytes are waiting in the queue to send to the USB host.
    //If any bytes are waiting, and the endpoint is available, prepare to
    //send the USB packet to the host.
    if ((USBUSARTIsTxTrfReady()) && (NextUSBOut > 0))
    {
        putUSBUSART(&USB_Out_Buffer[0], NextUSBOut);
        NextUSBOut = 0;
    }

    CDCTxService();
}//end ProcessIO

WORD_VAL ReadPOT(void)
{
    WORD_VAL w;

    w.Val = 0;

    ADCON0bits.GO = 1; // Start AD conversion
    while (ADCON0bits.GO); // Wait for conversion
    w.v[0] = ADRESL;
    w.v[1] = ADRESH;

    return w;
}

/********************************************************************
 * Function:        void BlinkUSBStatus(void)
 *
 * PreCondition:    None
 *
 * Input:           None
 *
 * Output:          None
 *
 * Side Effects:    None
 *
 * Overview:        BlinkUSBStatus turns on and off LEDs
 *                  corresponding to the USB device state.
 *
 * Note:            mLED macros can be found in HardwareProfile.h
 *                  USBDeviceState is declared and updated in
 *                  usb_device.c.
 *******************************************************************/
void BlinkUSBStatus(void)
{
    static WORD led_count = 0;

    if (led_count == 0)led_count = 10000U;
    led_count--;

    #define mLED_Both_Off()         {mLED_1_Off();mLED_2_Off();}
    #define mLED_Both_On()          {mLED_1_On();mLED_2_On();}
    #define mLED_Only_1_On()        {mLED_1_On();mLED_2_Off();}
    #define mLED_Only_2_On()        {mLED_1_Off();mLED_2_On();}

    if (USBSuspendControl == 1)
    {
        if (led_count == 0)
        {
            mLED_1_Toggle();
            if (mGetLED_1())
            {
                mLED_2_On();
            }
            else
            {
                mLED_2_Off();
            }
        }//end if
    }
    else
    {
        if (USBDeviceState == DETACHED_STATE)
        {
            mLED_Both_Off();
        }
        else if (USBDeviceState == ATTACHED_STATE)
        {
            mLED_Both_On();
        }
        else if (USBDeviceState == POWERED_STATE)
        {
            mLED_Only_1_On();
        }
        else if (USBDeviceState == DEFAULT_STATE)
        {
            mLED_Only_2_On();
        }
        else if (USBDeviceState == ADDRESS_STATE)
        {
            if (led_count == 0)
            {
                mLED_1_Toggle();
                mLED_2_Off();
            }//end if
        }
        else if (USBDeviceState == CONFIGURED_STATE)
        {
            if (led_count == 0)
            {
                mLED_1_Toggle();
                if (mGetLED_1())
                {
                    mLED_2_Off();
                }
                else
                {
                    mLED_2_On();
                }
            }//end if
        }
    }
}//end BlinkUSBStatus

void USBCBSuspend(void)
{
}

void USBCBWakeFromSuspend(void)
{
}

void USBCB_SOF_Handler(void)
{
}

void USBCBErrorHandler(void)
{
}

void USBCBCheckOtherReq(void)
{
    USBCheckHIDRequest();
    USBCheckCDCRequest();
}

void USBCBStdSetDscHandler(void)
{
}

void USBCBInitEP(void)
{
    //enable the HID endpoint
    USBEnableEndpoint(HID_EP, USB_IN_ENABLED | USB_OUT_ENABLED | USB_HANDSHAKE_ENABLED | USB_DISALLOW_SETUP);
    //Re-arm the OUT endpoint for the next packet
    USBOutHandle = HIDRxPacket(HID_EP, (BYTE*) & ReceivedDataBuffer, 64);

    CDCInitEP();
}

void USBCBSendResume(void)
{
    static WORD delay_count;

    //First verify that the host has armed us to perform remote wakeup.
    //It does this by sending a SET_FEATURE request to enable remote wakeup,
    //usually just before the host goes to standby mode (note: it will only
    //send this SET_FEATURE request if the configuration descriptor declares
    //the device as remote wakeup capable, AND, if the feature is enabled
    //on the host (ex: on Windows based hosts, in the device manager
    //properties page for the USB device, power management tab, the
    //"Allow this device to bring the computer out of standby." checkbox
    //should be checked).
    if (USBGetRemoteWakeupStatus() == TRUE)
    {
        //Verify that the USB bus is in fact suspended, before we send
        //remote wakeup signalling.
        if (USBIsBusSuspended() == TRUE)
        {
            USBMaskInterrupts();

            //Clock switch to settings consistent with normal USB operation.
            USBCBWakeFromSuspend();
            USBSuspendControl = 0;
            USBBusIsSuspended = FALSE; //So we don't execute this code again,
            //until a new suspend condition is detected.

            //Section 7.1.7.7 of the USB 2.0 specifications indicates a USB
            //device must continuously see 5ms+ of idle on the bus, before it sends
            //remote wakeup signalling.  One way to be certain that this parameter
            //gets met, is to add a 2ms+ blocking delay here (2ms plus at
            //least 3ms from bus idle to USBIsBusSuspended() == TRUE, yeilds
            //5ms+ total delay since start of idle).
            delay_count = 3600U;
            do
            {
                delay_count--;
            }
            while (delay_count);

            //Now drive the resume K-state signalling onto the USB bus.
            USBResumeControl = 1; // Start RESUME signaling
            delay_count = 1800U; // Set RESUME line for 1-13 ms
            do
            {
                delay_count--;
            }
            while (delay_count);
            USBResumeControl = 0; //Finished driving resume signalling

            USBUnmaskInterrupts();
        }
    }
}

#if defined(ENABLE_EP0_DATA_RECEIVED_CALLBACK)

void USBCBEP0DataReceived(void)
{
}
#endif

BOOL USER_USB_CALLBACK_EVENT_HANDLER(int event, void *pdata, WORD size)
{
    switch (event)
    {
        case EVENT_TRANSFER:
            //Add application specific callback task or callback function here if desired.
            break;
        case EVENT_SOF:
            USBCB_SOF_Handler();
            break;
        case EVENT_SUSPEND:
            USBCBSuspend();
            break;
        case EVENT_RESUME:
            USBCBWakeFromSuspend();
            break;
        case EVENT_CONFIGURED:
            USBCBInitEP();
            break;
        case EVENT_SET_DESCRIPTOR:
            USBCBStdSetDscHandler();
            break;
        case EVENT_EP0_REQUEST:
            USBCBCheckOtherReq();
            break;
        case EVENT_BUS_ERROR:
            USBCBErrorHandler();
            break;
        case EVENT_TRANSFER_TERMINATED:
            break;
        default:
            break;
    }
    return TRUE;
}

void InitializeUSART(void)
{
    unsigned char c;

    UART_TRISRx = 1; // RX
    UART_TRISTx = 0; // TX
    TXSTA = 0x24; // TX enable BRGH=1
    RCSTA = 0x90; // Single Character RX
    SPBRG = 0x71;
    SPBRGH = 0x02; // 0x0271 for 48MHz -> 19200 baud
    BAUDCON = 0x08; // BRG16 = 1
    c = RCREG; // read
}

void putcUSART(char c)
{
    TXREG = c;
}

unsigned char getcUSART(void)
{
    char c;

    if (RCSTAbits.OERR) // in case of overrun error
    { // we should never see an overrun error, but if we do,
        RCSTAbits.CREN = 0; // reset the port
        c = RCREG;
        RCSTAbits.CREN = 1; // and keep going.
    }
    else
    {
        c = RCREG;
    }

    return c;
}

#if defined(USB_CDC_SET_LINE_CODING_HANDLER)

void mySetLineCodingHandler(void)
{
    DWORD_VAL dwBaud;

    //If the request is not in a valid range
    if (cdc_notice.GetLineCoding.dwDTERate.Val > 115200)
    {
        //NOTE: There are two ways that an unsupported baud rate could be
        //handled.  The first is just to ignore the request and don't change
        //the values.  That is what is currently implemented in this function.
        //The second possible method is to stall the STATUS stage of the request.
        //STALLing the STATUS stage will cause an exception to be thrown in the
        //requesting application.  Some programs, like HyperTerminal, handle the
        //exception properly and give a pop-up box indicating that the request
        //settings are not valid.  Any application that does not handle the
        //exception correctly will likely crash when this requiest fails.  For
        //the sake of example the code required to STALL the status stage of the
        //request is provided below.  It has been left out so that this demo
        //does not cause applications without the required exception handling
        //to crash.
        //---------------------------------------
        //USBStallEndpoint(0,1);
    }
    else
    {
        //Update the baudrate info in the CDC driver
        CDCSetBaudRate(cdc_notice.GetLineCoding.dwDTERate.Val);

        dwBaud.Val = (DWORD) (GetSystemClock() / 4) / line_coding.dwDTERate.Val - 1;
        SPBRG = dwBaud.v[0];
        SPBRGH = dwBaud.v[1];
    }
}
#endif

char IdleI2C(void)
{
    unsigned long Timeout;

    Timeout = 10;
    // Idle;
    while (SSPCON2bits.ACKEN || SSPCON2bits.RCEN || SSPCON2bits.PEN || SSPCON2bits.RSEN || SEN || SSPSTATbits.R_nW)
    {
        if (Timeout-- == 0)
            return (-1);
    }
    return (0);
}

char StartI2C(void)
{
    unsigned long Timeout;

    Timeout = 100;
    // Start Condition
    SSPCON2bits.SEN = 1;
    while (SSPCON2bits.SEN)
    {
        if (Timeout-- == 0)
            return (-1);
    }
    return (0);
}

char StopI2C(void)
{
    unsigned long Timeout;

    Timeout = 100;
    // Stop Condition
    SSPCON2bits.PEN = 1;
    while (SSPCON2bits.PEN)
    {
        if (Timeout-- == 0)
            return (-1);
    }
    return (0);
}

char ReStartI2C(void)
{
    unsigned long Timeout;

    Timeout = 100;
    // ReStart Condition
    SSPCON2bits.RSEN = 1;
    while (SSPCON2bits.RSEN)
    {
        if (Timeout-- == 0)
            return (-1);
    }
    return (0);
}

char NAckI2C(void)
{
    unsigned long Timeout;

    Timeout = 100;
    // NACK Condition
    SSPCON2bits.ACKDT = 1;
    SSPCON2bits.ACKEN = 1;
    while (SSPCON2bits.ACKEN)
    {
        if (Timeout-- == 0)
            return (-1);
    }
    return (0);
}

// Master Mode Only

char WriteI2C(unsigned char Data)
{
    unsigned long Timeout;

    SSPBUF = Data;

    if (SSPCON1bits.WCOL)
    {
        return (-2);
    }

    else
    {
        Timeout = 100;
        while (SSPSTATbits.BF)
        {
            if (Timeout-- == 0)
                return (-1);
        }
        if (IdleI2C())
            return (-3);
        if (SSPCON2bits.ACKSTAT)
            return (-4);
        else return (0);
    }
}

// Master Mode Only

char ReadI2C(unsigned char * Data)
{
    unsigned long Timeout;

    Timeout = 100;
    SSP1CON2bits.RCEN = 1;
    while (SSPCON2bits.RCEN)
    {
        if (Timeout-- == 0)
            return (-1);
    }

    Timeout = 10;
    while (!SSP1STATbits.BF)
    {
        if (Timeout-- == 0)
            return (-1);
    }

    * Data = SSP1BUF;

    return (0);
}

char I2C_EE16BitsByteWirte(unsigned char SalveAddr, unsigned int Addr, unsigned char Data)
{
    if (IdleI2C())
        return (-1);

    if (StartI2C())
        return (-1);

    if (WriteI2C(SalveAddr & 0xfe))
        return (-1);

    if (WriteI2C((Addr >> 8) & 0xff))
        return (-1);

    if (WriteI2C(Addr & 0xff))
        return (-1);

    if (WriteI2C(Data))
        return (-1);

    if (StopI2C())
        return (-1);

    return (0);
}

char I2C_EE16BitsRandomRead(unsigned char SalveAddr, unsigned int Addr, unsigned char * Data)
{
    if (IdleI2C())
        return (-1);

    if (StartI2C())
        return (-1);

    if (WriteI2C(SalveAddr & 0xfe))
        return (-1);

    if (WriteI2C((Addr >> 8) & 0xff))
        return (-1);

    if (WriteI2C(Addr & 0xff))
        return (-1);

    if (ReStartI2C())
        return (-1);

    if (WriteI2C(SalveAddr | 0x01))
        return (-1);

    if (ReadI2C(Data))
        return (-1);

    if (NAckI2C())
        return (-1);

    if (StopI2C())
        return (-1);

    return (0);
}
